#!/usr/bin/perl
#       Copyright Hunter Moseley, 2009. All rights reserved.
my $help = <<HELP;
analyze_solid_state.pl - perform a bottom up grouping and tuple-based typing of spin systems using a  set of solid-state peak lists.
   The peak lists are all in one AutoAssign formated peak list file, identified by the following workbook names:
      CANCOCX - CAi, Ni, COi-1, CXi-1
      CANCOCO - CAi, Ni, COi-1, CO(i,i-1) ; trivial subset of CANCOCX.
      CANcoCA - CAi, Ni, CAi-1
      NCACX   - CAi, Ni, CXi
   The program generates the following set of Sparky formated peak lists ready for input into AutoAssign.
      HNCO.list
      HNcaCO.list
      HNCACB.list
      HNcoCACB.list
   An artificial HN chemical shift has been created from the intra CA shift divided by 6.
   Assignment constraints have been added to the HNCO.list file to limit intra and sequential amino acid typing.

 usage: analyze_solid_state all.pks protein_sequence [options]

 Parameters:
	all.pks - AutoAssign formatted peak list with peaks from different experiments identified by their workbook name.
	protein_sequence - single letter amino acid sequence of protein.
        options:
		-NCACX CA_tol N_tol - matching tolerances for grouping NCACX peaks.
		-CANCOCX CA_tol N_tol CO_tol - matching tolerances for grouping CANCOCX peaks.
		-CANcoCA_NCACX - CA_tol N_tol CX_tol - matching tolerances for grouping NCACX peak groups to CANcoCA peaks.
		-CANcoCA_CANCOCX - CA_tol N_tol CX_tol - matching tolerances for grouping CANCOCX peak groups to CANcoCA peaks.


   All match tolerances should be between 2 to 4 standard deviations of dimension registrations calculated using calculate_registration.
HELP


use FindBin;
use lib $FindBin::Bin;

use Statistics::Distributions;

# Load needed Modules for Bayesian statistics use BMRB::ChemicalShift;
use BMRB::BayesianCalculations2 qw(:ALL);
use BMRB::BayesianCalculations3 qw(:ALL);
use BMRB::BayesianCalculations4 qw(:ALL);
use BMRB::ExpectedValues qw(:ALL);

use strict;

#
# Parse command line arguments
#
if (! @ARGV)
  {
  print $help;
  exit;
  }

# Read in comand line parameters
my $command_line = $0 . " " . join(' ',@ARGV);
my $in_peak_file = shift @ARGV;
my $protein_sequence = uc(shift @ARGV);
my @NCACX_NCACX_tolerances = (0.15, 0.33);
my @CANcoCA_NCACX_tolerances = (0.3, 0.6); #(0.2, 0.45);
my @CANCOCX_CANCOCX_tolerances = (1, 0.28, 0.2);
my @CANcoCA_CANCOCX_tolerances = (1, 0.8, 0.0, 0.5);

while(@ARGV && $ARGV[0] =~ /^-/)
  {
  my $switch = shift @ARGV;
  if ($switch =~ /^-NCACX/)
    { @NCACX_NCACX_tolerances = (shift @ARGV, shift @ARGV); }
  elsif ($switch =~ /^-CANcoCA_NCACX/)
    { @CANcoCA_NCACX_tolerances = (shift @ARGV, shift @ARGV); }
  elsif ($switch =~ /^-CANCOCX/)
    { @CANCOCX_CANCOCX_tolerances = (shift @ARGV, shift @ARGV, shift @ARGV); }
  elsif ($switch =~ /^-CANcoCA_CANCOCX/)
    { @CANcoCA_CANCOCX_tolerances = (shift @ARGV, shift @ARGV, 0, shift @ARGV); }
  else
    {
    print STDERR "Error: unknown option ",$switch," given.\n";
    exit(1);
    }
  }



#
# Definitions for subroutines
#
my $ignore_peak_based_overrides = 0;
my %linear_mapping = (map { ("$_" => $_); } (1..100));

# Initialize amino acid statistics for amino acid typing
my ($mean_count, $mean, $variance, $std, $min_value, $max_value) = &amino_acid_expected_shifts(0);
my $restricted_type_lists = &amino_acid_type_lists(1,1,1);
my %expected_aa_hash;
foreach my $aa (split(//,$protein_sequence))
  { $expected_aa_hash{$aa}++; }
my @expected_aa = (keys %expected_aa_hash);
print "Expected AA = ",join(" ", map { $_ . " = " . $expected_aa_hash{$_}; } (@expected_aa)),"\n";

# Build CA range for testing.
my %CA_range;
foreach my $aa (@expected_aa)
  { $CA_range{$aa} = [ (($$min_value{$aa}{"CA"} > ($$mean{$aa}{"CA"} - 4 * $$std{$aa}{"CA"})) ? $$min_value{$aa}{"CA"} : ($$mean{$aa}{"CA"} - 4 * $$std{$aa}{"CA"})), (($$max_value{$aa}{"CA"} < ($$mean{$aa}{"CA"} + 4 * $$std{$aa}{"CA"})) ? $$max_value{$aa}{"CA"} : ($$mean{$aa}{"CA"} + 4 * $$std{$aa}{"CA"})) ]; }

# Build CO range for testing.
my %CO_range;
foreach my $aa (@expected_aa)
  { $CO_range{$aa} = [ (($$min_value{$aa}{"C"} > ($$mean{$aa}{"C"} - 4 * $$std{$aa}{"C"})) ? $$min_value{$aa}{"C"} : ($$mean{$aa}{"C"} - 4 * $$std{$aa}{"C"})), (($$max_value{$aa}{"C"} < ($$mean{$aa}{"C"} + 4 * $$std{$aa}{"C"})) ? $$max_value{$aa}{"C"} : ($$mean{$aa}{"C"} + 4 * $$std{$aa}{"C"})) ]; }


# Build CB range for testing.
my %CB_range;
foreach my $aa (grep { exists $$min_value{$_}{"CB"}; } (@expected_aa))
  { $CB_range{$aa} = [ (($$min_value{$aa}{"CB"} > ($$mean{$aa}{"CB"} - 4 * $$std{$aa}{"CB"})) ? $$min_value{$aa}{"CB"} : ($$mean{$aa}{"CB"} - 4 * $$std{$aa}{"CB"})), (($$max_value{$aa}{"CB"} < ($$mean{$aa}{"CB"} + 4 * $$std{$aa}{"CB"})) ? $$max_value{$aa}{"CB"} : ($$mean{$aa}{"CB"} + 4 * $$std{$aa}{"CB"})) ]; }


# preparation to calculate PRTL and shift assignments.
my %PRTL_props;
$PRTL_props{"mean"} = $mean;
$PRTL_props{"mean_count"} = $mean_count;
$PRTL_props{"variance"} = $variance;
$PRTL_props{"File_ResType_frequencies"}{"TEST"} = \%expected_aa_hash;
$PRTL_props{"restricted_type_lists"} = $restricted_type_lists;
$PRTL_props{"covariance_type_lists"} = $restricted_type_lists;
$PRTL_props{"single_bp_sub"} = \&singleBProbByMeanVarianceChiSquare; 
$PRTL_props{"prior_sub"} = \&simplePrior; 
$PRTL_props{"CX_prior_sub"} = \&simplePrior; 
$PRTL_props{"Shift_Count_Prior_Weight"} = 5; 


#
# Begining of main execution
#


# Read Peak file
my ($peak_alist, $end_comments_alist) = &read_peak_file($in_peak_file);

# remove duplicates
my %dup_count;
my @clean_alist;
foreach my $peak (@$peak_alist)
  {
  my $test_string = join(".", map { $$peak[$_]; } (1..($#$peak-2)));
  next if ($dup_count{$test_string}++);
  push @clean_alist, $peak;
  }

# remove negative peaks
@clean_alist = grep { ($$_[$#$_-1] > 0); } (@clean_alist);

# separate into known lists
my @CANCOCX = grep { $$_[$#$_] =~ /^CANCOCX(\.[a-zA-Z0-9_;]+)?$/; } (@clean_alist);
my @CANCOCO = grep { $$_[$#$_] =~ /^CANCOCO(\.[a-zA-Z0-9_;]+)?$/; } (@clean_alist);
my @CANcoCA = grep { $$_[$#$_] =~ /^CANcoCA(\.[a-zA-Z0-9_;]+)?$/; } (@clean_alist);
my @NCACX = grep { $$_[$#$_] =~ /^NCACX(\.[a-zA-Z0-9_;]+)?$/; } (@clean_alist);

print "Orig peak list = ",scalar(@$peak_alist),"\n";
print "Without negative and duplicate peaks = ",scalar(@clean_alist),"\n";

print "CANCOCX = ",scalar(@CANCOCX),"\n";
print "CANCOCO = ",scalar(@CANCOCO),"\n";
print "CANcoCA = ",scalar(@CANcoCA),"\n";
print "NCACX = ",scalar(@NCACX),"\n";

# remove peaks that do not look like CA.
@CANcoCA = grep { &verify_shift($$_[1],\%CA_range) && &verify_shift($$_[3],\%CA_range); } (@CANcoCA);
print "tested CANcoCA = ",scalar(@CANcoCA),"\n";


# group NCACX peaks together
my @NCACX_groups = ({ "peaks" => [ $NCACX[0] ] });
foreach my $peak (map { $NCACX[$_]; } (1..$#NCACX))
  {
  # find peak_groups that the peak should be added to
  my @add_list = grep { &valid_grouping($peak,$NCACX_groups[$_]{"peaks"}, \@NCACX_NCACX_tolerances); } (0..$#NCACX_groups);
  
  if (! @add_list) # create new peak_group because no suitable group was found.
    {
    push @NCACX_groups, { "peaks" => [ $peak ] };
    next;
    }

  if (@add_list > 1) # if there are more than one peak_group, then combine them.
    {
    my @hold_list;
    @add_list = sort { $b <=> $a; } @add_list; 
    foreach my $index (@add_list)
      {
      my $temp_group = splice(@NCACX_groups, $index, 1);
      push @hold_list, @{$$temp_group{"peaks"}}; 
      }
    
    push @NCACX_groups, { "peaks" => [ @hold_list ] };
    @add_list = ($#NCACX_groups);
    }
  
  push @{$NCACX_groups[$add_list[0]]{"peaks"}}, $peak;
  }

print "NCACX groups = ",scalar(@NCACX_groups),"\n";



# calculate closest match of CANcoCA to each NCACX group
#  This section incomplete, but good enough for my immediate purposes.
my @distance_list = ();
my $count = 0;
foreach my $peak (@CANcoCA)
  {
  my @close_list = grep { &valid_grouping($peak,$NCACX_groups[$_]{"peaks"}, \@CANcoCA_NCACX_tolerances); } (0..$#NCACX_groups);
  foreach my $group (map { $NCACX_groups[$_]; } (@close_list))
    { push @distance_list, [ &calc_distance($peak, $$group{"peaks"}, \@CANcoCA_NCACX_tolerances), $peak, $group ]; }
  if (@close_list)
    { $count++; }
  }

print "Within Tolerance CANcoCA = $count\n";

foreach my $distance (sort { $$a[0] <=> $$b[0]; } (@distance_list))
  {
  if (! exists $$distance[2]{"CANcoCA"})
    { $$distance[2]{"CANcoCA"} = $$distance[1]; }
  }

my @matched_NCACX_groups = grep { exists $$_{"CANcoCA"}; } (@NCACX_groups);

print "Matched CANcoCA = ",scalar(@matched_NCACX_groups),"\n";
print "Distances = ", join(" ", sort { $b <=> $a; } (map { &calc_distance($$_{"CANcoCA"}, $$_{"peaks"}, \@CANcoCA_NCACX_tolerances); } (@matched_NCACX_groups))),"\n";
print "Unmatched group sizes = ", join("",map { scalar(@{$$_{"peaks"}}) . " "; } (grep { ! exists $$_{"CANcoCA"} && @{$$_{"peaks"}} > 1; } (@NCACX_groups))),"\n";



# group CANCOX peaks together
my @CANCOCX_groups = ( { "peaks" => [ $CANCOCX[0] ] } );
foreach my $peak (map { $CANCOCX[$_]; } (1..$#CANCOCX))
  {
  # find peak_groups that the peak should be added to
  my @add_list = grep { &valid_grouping($peak,$CANCOCX_groups[$_]{"peaks"}, \@CANCOCX_CANCOCX_tolerances); } (0..$#CANCOCX_groups);
  
  if (! @add_list) # create new peak_group because no suitable group was found.
    {
    push @CANCOCX_groups, { "peaks" => [ $peak ] };
    next;
    }

  if (@add_list > 1) # if there are more than one peak_group, then combine them.
    {
    my @hold_list;
    @add_list = sort { $b <=> $a; } @add_list; 
    foreach my $index (@add_list)
      {
      my $temp_group = splice(@CANCOCX_groups, $index, 1);
      push @hold_list, @{$$temp_group{"peaks"}}; 
      }
    
    push @CANCOCX_groups, { "peaks" => [ @hold_list ] };
    @add_list = ($#CANCOCX_groups);
    }
  
  push @{$CANCOCX_groups[$add_list[0]]{"peaks"}}, $peak;
  }

print "CANCOCX groups = ",scalar(@CANCOCX_groups),"\n";

print " group sizes = ", join("",map { scalar(@{$$_{"peaks"}}) . " "; } (@CANCOCX_groups)),"\n";

# calculate closest match of CANcoCA to each CANCOCX group
#  This section incomplete, but good enough for my immediate purposes.
my %CANcoCA_CANCOCX_mapping = ("1" => 1, "2" => 2, "4" => 3);
my @distance_list = ();
my $count = 0;
foreach my $peak (map { $$_{"CANcoCA"}; } (@matched_NCACX_groups))
  {
  my @close_list = grep { &valid_grouping($peak,$CANCOCX_groups[$_]{"peaks"}, \@CANcoCA_CANCOCX_tolerances, \%CANcoCA_CANCOCX_mapping); } (0..$#CANCOCX_groups);
  foreach my $group (map { $CANCOCX_groups[$_]; } (@close_list))
    { push @distance_list, [ &calc_distance($peak, $$group{"peaks"}, \@CANcoCA_CANCOCX_tolerances, \%CANcoCA_CANCOCX_mapping), $peak, $group ]; }
  if (@close_list)
    { $count++; }
  }

print "Within Tolerance CANcoCA = $count\n";

foreach my $distance (sort { $$a[0] <=> $$b[0]; } (@distance_list))
  {
  if (! exists $$distance[2]{"CANcoCA"})
    { $$distance[2]{"CANcoCA"} = $$distance[1]; }
  }

my @matched_CANCOCX_groups = grep { exists $$_{"CANcoCA"}; } (@CANCOCX_groups);

print "Matched CANcoCA = ",scalar(@matched_CANCOCX_groups),"\n";
print "Distances = ", join(" ", sort { $b <=> $a; } (map { &calc_distance($$_{"CANcoCA"}, $$_{"peaks"}, \@CANcoCA_CANCOCX_tolerances, \%CANcoCA_CANCOCX_mapping); } (@matched_CANCOCX_groups))),"\n";
print "Unmatched group sizes = ", join("",map { scalar(@{$$_{"peaks"}}) . " "; } (grep { ! exists $$_{"CANcoCA"} && @{$$_{"peaks"}} > 1; } (@CANCOCX_groups))),"\n";


# create mapping hash to relate CANcoCA to matched NCACX groups.
my %CANcoCA_NCACX_group_map = (map { ($$_{"CANcoCA"} => $_);} (@matched_NCACX_groups));

my @fully_matched_CANCOCX_groups = grep { exists $CANcoCA_NCACX_group_map{$$_{"CANcoCA"}}; } (@matched_CANCOCX_groups);


# assign directly determinable shifts
foreach my $group (@fully_matched_CANCOCX_groups)
  {
  my $NCACX_group = $$group{"NCACX_group"} = $CANcoCA_NCACX_group_map{$$group{"CANcoCA"}};
  my $used_peaks = $$group{"used_peaks"} = {};
  

  # assign CA and N shifts
  $$group{"intra_shifts"}{"CA"} = $$group{"CANcoCA"}[1];
  $$group{"intra_shifts"}{"N"} = $$group{"CANcoCA"}[2];
  $$group{"seq_shifts"}{"CA"} = $$group{"CANcoCA"}[3];

  # remove intra CA peaks from intras
  foreach my $peak (grep { abs($$_[1] - $$_[3]) <= 0.2; } (@{$$NCACX_group{"peaks"}}))
    { $$used_peaks{$peak}++; }


  # remove seq CA peaks from seqs
  foreach my $peak (grep { abs($$group{"seq_shifts"}{"CA"} - $$_[4]) <= 0.3; } (@{$$group{"peaks"}}))
    { $$used_peaks{$peak}++; }

  
  # assign seq CO shift if unique
  my @possible_seq_CO = grep { &verify_shift($$_[3],\%CO_range); } (@{$$group{"peaks"}});
  if (@possible_seq_CO && ! grep {abs($possible_seq_CO[0][3] - $$_[3]) > 0.3; } (@possible_seq_CO))
    {
    $$group{"seq_shifts"}{"C"} = $possible_seq_CO[0][3];
    @possible_seq_CO = ();
    }

  # assign intra CO shift if unique.
#  my @possible_intra_CO = sort { $$b[4] <=> $$a[4]; } (grep { &verify_shift($$_[3],\%CO_range); } (@{$$NCACX_group{"peaks"}}));
#  if (exists $$group{"seq_shifts"}{"C"} && @possible_intra_CO > 1) # remove sequential CO from intras.
#    {
#    foreach my $peak (grep { abs($$group{"seq_shifts"}{"C"} - $$_[3]) <= 0.3; } (@possible_intra_CO))
#      { $$used_peaks{$peak}++; }
#    
#    @possible_intra_CO = grep { abs($$group{"seq_shifts"}{"C"} - $$_[3]) > 0.3; } (@possible_intra_CO);
#    }

#  if (@possible_intra_CO && ! grep {abs($possible_intra_CO[0][3] - $$_[3]) > 0.3; } (@possible_intra_CO))
#    {
#    foreach my $peak (@possible_intra_CO)
#      { $$used_peaks{$peak}++; }
#    $$group{"intra_shifts"}{"C"} = $possible_intra_CO[0][3];
#    @possible_intra_CO = ();
#    }

  }

# try to assign the rest of the C shifts
foreach my $group (@fully_matched_CANCOCX_groups)
  {
  my $NCACX_group = $$group{"NCACX_group"};
  my $used_peaks = $$group{"used_peaks"};
  $$NCACX_group{"unused_peaks"} = [ grep { ! exists $$used_peaks{$_}; } (@{$$NCACX_group{"peaks"}}) ];
  $$group{"unused_peaks"} = [ grep { ! exists $$used_peaks{$_}; } (@{$$group{"peaks"}}) ];


  if (! exists $$group{"intra_shifts"}{"C"}) # try to assign intra C shifts
    {
    my %verified_peaks = ();
    my $min_shifts = 0;

    foreach my $group2 (grep { $_ != $group && abs($$_{"seq_shifts"}{"CA"} - $$group{"intra_shifts"}{"CA"}) < 0.5; } (@fully_matched_CANCOCX_groups))
      {
      my @verified = grep { abs($$_[3] - $$group2{"seq_shifts"}{"C"}) < 0.5; } (@{$$NCACX_group{"unused_peaks"}});

      my $test_min=0;
      foreach my $peak (@verified)
	{ $test_min += 1.0 / scalar(grep { abs($$_[3]- $$peak[3]) < 0.3; } (@verified)); }
      if (int($test_min+0.5) > $min_shifts)
	{ $min_shifts = int($test_min+0.5); }
      foreach my $peak (@verified)
	{ $verified_peaks{$peak} = $peak; }
      }

    # min_shifts may need to be 1 less in the full algorithm in order to follow the 1-error rule.
    if (%verified_peaks)
      { &PRTL_with_shift_assignments($$group{"intra_shifts"}, [ values %verified_peaks ], 3, \%PRTL_props, 0.9999, 0.2, $min_shifts,$used_peaks,"C", [ "C" ], \%CO_range); }
    elsif (@{$$NCACX_group{"unused_peaks"}})
      { &PRTL_with_shift_assignments($$group{"intra_shifts"}, [ grep { ! exists $$group{"seq_shifts"}{"C"} || (abs($$group{"seq_shifts"}{"C"} - $$_[3]) > 0.3);  } (@{$$NCACX_group{"unused_peaks"}}) ], 3, \%PRTL_props, 0.9999, 0.2, 1, $used_peaks,"C", [ "C" ], \%CO_range); }
    }


  if (! exists $$group{"seq_shifts"}{"C"}) # try to assign seq C shifts
    {
    my %verified_peaks = ();
    my $min_shifts = 0;

    foreach my $group2 (grep { $_ != $group && abs($$_{"intra_shifts"}{"CA"} - $$group{"seq_shifts"}{"CA"}) < 0.5; } (@fully_matched_CANCOCX_groups))
      {
      my @verified = grep { abs($$_[4] - $$group2{"intra_shifts"}{"C"}) < 0.5; } (@{$$group{"unused_peaks"}});

      my $test_min=0;
      foreach my $peak (@verified)
	{ $test_min += 1.0 / scalar(grep { abs($$_[4]- $$peak[4]) < 0.3; } (@verified)); }
      if (int($test_min+0.5) > $min_shifts)
	{ $min_shifts = int($test_min+0.5); }
      foreach my $peak (@verified)
	{ $verified_peaks{$peak} = $peak; }
      }

    # min_shifts may need to be 1 less in the full algorithm in order to follow the 1-error rule.
    if (%verified_peaks)
      { &PRTL_with_shift_assignments($$group{"seq_shifts"}, [ values %verified_peaks ], 4, \%PRTL_props, 0.9999, 0.3, $min_shifts,$used_peaks,"C", [ "C" ], \%CO_range); }
    }
  }

foreach my $group (@fully_matched_CANCOCX_groups)
  {
  my $NCACX_group = $$group{"NCACX_group"};
  my $used_peaks = $$group{"used_peaks"};
  $$NCACX_group{"unused_peaks"} = [ grep { ! exists $$used_peaks{$_}; } (@{$$NCACX_group{"peaks"}}) ];
  $$group{"unused_peaks"} = [ grep { ! exists $$used_peaks{$_}; } (@{$$group{"peaks"}}) ];

  
  # determine other intra shifts
  if (! exists $$group{"intra_PRTL"})
    {    
    my %verified_peaks = ();
    my $min_shifts = 0;

    foreach my $group2 (grep { $_ != $group && (abs($$_{"seq_shifts"}{"CA"} - $$group{"intra_shifts"}{"CA"}) < 0.5 || (exists $$_{"seq_shifts"}{"C"} && exists $$group{"intra_shifts"}{"C"} && abs($$_{"seq_shifts"}{"C"} - $$group{"intra_shifts"}{"C"}) < 0.5)); } (@fully_matched_CANCOCX_groups))
      {
      my @verified = grep { my $test = $_; (grep { abs($$test[3] - $$_[4]) < 0.5; } (@{$$group2{"unused_peaks"}})); } (@{$$NCACX_group{"unused_peaks"}});
      if (! exists $$group{"intra_shifts"}{"C"} && exists $$group2{"seq_shifts"}{"C"})
	{ push @verified, grep { abs($$_[3] - $$group2{"seq_shifts"}{"C"}) < 0.5; } (@{$$NCACX_group{"unused_peaks"}}); }

      my $test_min=0;
      foreach my $peak (@verified)
	{ $test_min += 1.0 / scalar(grep { abs($$_[3]- $$peak[3]) < 0.3; } (@verified)); }
      if (int($test_min+0.5) > $min_shifts)
	{ $min_shifts = int($test_min+0.5); }
      foreach my $peak (@verified)
	{ $verified_peaks{$peak} = $peak; }
      }

    # min_shifts may need to be 1 less in the full algorithm in order to follow the 1-error rule.
    if (%verified_peaks)
      { $$group{"intra_PRTL"} = &PRTL_with_shift_assignments($$group{"intra_shifts"}, [ values %verified_peaks ], 3, \%PRTL_props, 0.9999, 0.2, $min_shifts,$used_peaks,"CB", [ "CB" ], \%CB_range); }
    elsif (@{$$NCACX_group{"unused_peaks"}})
      { $$group{"intra_PRTL"} = &PRTL_with_shift_assignments($$group{"intra_shifts"}, $$NCACX_group{"unused_peaks"}, 3, \%PRTL_props, 0.9999, 0.2, 1,$used_peaks,"CB", [ "CB" ], \%CB_range); }
    }


  # determine other seq shifts
  if (! exists $$group{"seq_PRTL"})
    {
    my %verified_peaks = ();
    my $min_shifts = 0;

    foreach my $group2 (grep { $_ != $group && (abs($$_{"intra_shifts"}{"CA"} - $$group{"seq_shifts"}{"CA"}) < 0.5 || (exists $$_{"intra_shifts"}{"C"} && exists $$group{"seq_shifts"}{"C"} && abs($$_{"intra_shifts"}{"C"} - $$group{"seq_shifts"}{"C"}) < 0.5)); } (@fully_matched_CANCOCX_groups))
      {
      my @verified = grep { my $test = $_; (grep { abs($$test[4] - $$_[3]) < 0.5; } (@{$$group2{"NCACX_group"}{"unused_peaks"}})); } (@{$$group{"unused_peaks"}});
      if (! exists $$group{"seq_shifts"}{"C"} && exists $$group2{"intra_shifts"}{"C"})
	{ push @verified, grep { abs($$_[4] - $$group2{"intra_shifts"}{"C"}) < 0.5; } (@{$$group{"unused_peaks"}}); }

      my $test_min=0;
      foreach my $peak (@verified)
	{ $test_min += 1.0 / scalar(grep { abs($$_[4]- $$peak[4]) < 0.3; } (@verified)); }
      if (int($test_min+0.5) > $min_shifts)
	{ $min_shifts = int($test_min+0.5); }
      foreach my $peak (@verified)
	{ $verified_peaks{$peak} = $peak; }
      }

    # min_shifts may need to be 1 less in the full algorithm in order to follow the 1-error rule.
    if (%verified_peaks)
      { $$group{"seq_PRTL"} = &PRTL_with_shift_assignments($$group{"seq_shifts"}, [ values %verified_peaks ], 4, \%PRTL_props, 0.9999, 0.3, $min_shifts,$used_peaks,"CB", [ "CB" ], \%CB_range); }
    elsif (@{$$NCACX_group{"unused_peaks"}})
      { $$group{"seq_PRTL"} = &PRTL_with_shift_assignments($$group{"seq_shifts"}, $$group{"unused_peaks"}, 4, \%PRTL_props, 0.9999, 0.3, 1,$used_peaks,"CB", [ "CB" ], \%CB_range); }
    }

  # test for extra intra glycine
  if (! exists $$group{"intra_PRTL"} && $$group{"intra_shifts"}{"CA"} >= 39 && $$group{"intra_shifts"}{"CA"} <= 49 && grep { abs($$_[1] - $$_[3]) <= 0.2; } (@{$$NCACX_group{"peaks"}}))
    { 
    $$group{"intra_PRTL"} = [ "G" ]; 
    }

  # test for extra seq glycine
  if (! exists $$group{"seq_PRTL"} && $$group{"seq_shifts"}{"CA"} >= 39 && $$group{"seq_shifts"}{"CA"} <= 49 && grep { abs($$group{"seq_shifts"}{"CA"} - $$_[4]) <= 0.3; } (@{$$group{"peaks"}}))
    { 
    $$group{"seq_PRTL"} = [ "G" ]; 
    }
  }

print "intra C assigned = ",scalar(grep { exists $$_{"intra_shifts"}{"C"}; } (@fully_matched_CANCOCX_groups)),"\n";
print "seq C assigned = ",scalar(grep { exists $$_{"seq_shifts"}{"C"}; } (@fully_matched_CANCOCX_groups)),"\n";
print "intra CB assigned = ",scalar(grep { exists $$_{"intra_shifts"}{"CB"}; } (@fully_matched_CANCOCX_groups)),"\n";
print "seq CB assigned = ",scalar(grep { exists $$_{"seq_shifts"}{"CB"}; } (@fully_matched_CANCOCX_groups)),"\n";
my $average_intra_PRTL = 0;
my $count = 0;
foreach my $group (@fully_matched_CANCOCX_groups)
  {
  if (exists $$group{"intra_PRTL"})
    {
    $average_intra_PRTL += scalar(@{$$group{"intra_PRTL"}});
    $count++;
    }
  }
print "average intra PRTL = ",$average_intra_PRTL/$count,"\n";
my $average_seq_PRTL = 0;
my $count = 0;
foreach my $group (@fully_matched_CANCOCX_groups)
  {
  if (exists $$group{"seq_PRTL"})
    {
    $average_seq_PRTL += scalar(@{$$group{"seq_PRTL"}});
    $count++;
    }
  }
print "average seq PRTL = ",$average_seq_PRTL/$count,"\n";


#
# Create AutoAssign dataset
#
 
foreach my $group (@fully_matched_CANCOCX_groups)
  {
  # create all_shifts hash
  my $all_shifts = $$group{"all_shifts"} = {};
  foreach my $shift (keys %{$$group{"intra_shifts"}})
    { $$all_shifts{$shift} = $$group{"intra_shifts"}{$shift}; }
  foreach my $shift (keys %{$$group{"seq_shifts"}})
    { $$all_shifts{$shift . "-1"} = $$group{"seq_shifts"}{$shift}; }

  # create H shift
  $$all_shifts{"H"} = $$all_shifts{"CA"} / 6.0;
  }


my %pl_definitions;
$pl_definitions{"HNCO"} = [ [ 1, ["H", "N", "C-1" ] ] ];
$pl_definitions{"HNcaCO"} = [ [ 1, ["H", "N", "C" ] ] ];
$pl_definitions{"HNCACB"} = [ [ 1, ["H", "N", "CA" ] ], [ -1, ["H", "N", "CB" ] ] ];
$pl_definitions{"HNcoCACB"} = [ [ 1, ["H", "N", "CA-1" ] ], [ -1, ["H", "N", "CB-1" ] ] ];

foreach my $pl_type (keys %pl_definitions)
  {
  my $outfile = $pl_type . ".list";
  local *SPARKY_FILE;
  print "creating $outfile\n";
  open(SPARKY_FILE, ">$outfile") || die ("open " . $outfile . " failed.");
  
  foreach my $group (@fully_matched_CANCOCX_groups)
    {
    my $TAC = "";
    if ($pl_type eq "HNCO") 
      {
      if (exists $$group{"intra_PRTL"})
	{ $TAC .= ";i" . join("",@{$$group{"intra_PRTL"}}); }
      if (exists $$group{"seq_PRTL"})
	{ $TAC .= ";s" . join("",@{$$group{"seq_PRTL"}}); }
      }
      
    foreach my $peak_def (@{$pl_definitions{$pl_type}})
      {
      if (! grep { ! exists $$group{"all_shifts"}{$_}; } (@{$$peak_def[1]}))
	{ print SPARKY_FILE join("  ", map { $$group{"all_shifts"}{$_} + (rand() * 0.00001); } (@{$$peak_def[1]}))," ",$$peak_def[0], "  $TAC\n"; }
#	{ print SPARKY_FILE join("-", map { "?"; } (@{$$peak_def[1]})),"   ", join("  ", map { $$group{"all_shifts"}{$_} + (rand() * 0.00001); } (@{$$peak_def[1]}))," ",$$peak_def[0], "  $TAC\n"; }
      }

    }

  close SPARKY_FILE;
  }


#
# Subroutines
#

# PRTL_size
#   Creates and returns a PRTL by calling the appropriate Bayesian Typing method
#
# Parameters:
#	$res_shift_hlist - hash of shift_name to ChemicalShift
#	$PRTL_props - hash of properties used by Bayesian Typing methods
#	$cutoff - probability cutoff to use.
#
sub PRTL_with_shift_assignments
  {
  my $known_shifts = shift @_;
  my $peaks = shift @_;
  my $peak_index = shift @_;
  my $PRTL_props = shift @_;
  my $cutoff = shift @_;
  my $tolerance = shift @_;
  my $min_size = shift @_;
  my $used = shift @_;
  my $assign_type = shift @_;
  my $possible_types = shift @_;
  my $assign_range = shift @_;  

  # create res_shift_hlist
  my $res_shift_hlist = shift @_;  

  my $carbon_known = 0;
  foreach my $carbon_shift (grep { $_ =~ /^C/; } (keys %$known_shifts))
    { 
    $$res_shift_hlist{$carbon_shift} = BMRB::ChemicalShift->new( [ "TEST", 0, "X", $carbon_shift, $$known_shifts{$carbon_shift} ] ); 
    $carbon_known++;
    }


  my $count=1;
  foreach my $peak (@$peaks)
    { 
    $$res_shift_hlist{"CX" . $count} = BMRB::ChemicalShift->new( [ "TEST", $peak, "X", "CX" . $count, $$peak[$peak_index] ] ); 
    $count++;
    }

  print "shifts = ",join("  ", map { $_ . " " . $$res_shift_hlist{$_}->Value; } (keys %$res_shift_hlist)),"\n";

  my @prob_list = ();
  my $total_prob = 0.0;
  
  # creates an array of arrays where each row is the aminoacid type and the its probability
  foreach my $aa_type (@expected_aa)
    {
    my $used_res_shift_hlist;
    
    my ($prob_value, $assigned_shift_hlist) = &allBayesianProbCombinationsCXrecomb3($aa_type, $res_shift_hlist, $PRTL_props, 0, $min_size);
    push @prob_list, [ $aa_type, $prob_value, $assigned_shift_hlist ];
    $total_prob += $prob_value; 
    }

  
  # normalize each probablilty by the total
  if ($total_prob != 0.0)
    {
    foreach my $prob (@prob_list)
      { $$prob[1] = $$prob[1] / $total_prob; } 
    }
  
  # return the probabilites in descending order
  @prob_list = (sort { $$b[1] <=> $$a[1] } @prob_list);

  # create PRTL
  my $sum = 0;
  my $PRTL = [ map { $$_[0]; } (grep {my $test = ($sum <= $cutoff); $sum += $$_[1]; $test; } (@prob_list)) ];
  print "PRTL = ",join(" ",@$PRTL),"\n";

  # Try to assign CB
  foreach my $test_type (@$possible_types)
    {
    my ($shift_prob, $shift)  = &highest_shift_prob($test_type,\@prob_list,$tolerance);
    print "$test_type prob = ",$shift_prob,"\n";
    if (($shift_prob > 0.25) && &verify_shift($shift->Value, $assign_range))
      { 
      print "$assign_type = ",$shift->Value,"\n";
      $$known_shifts{$assign_type} = $shift->Value; 
      $$used{$shift->ResNum}++;
      last;
      }
    }

  return $PRTL;
  }


#
#
#
sub highest_shift_prob
  {
  my $shift_type = shift @_;
  my $prob_list = shift @_;
  my $tolerance = shift @_;

  my @shift_prob;
  foreach my $prob (@$prob_list)
    {
    my $sum = 0;
    if (exists $$prob[2]{$shift_type})
      {
      foreach my $prob2 (@$prob_list)
	{
	if (exists $$prob2[2]{$shift_type} && (abs($$prob[2]{$shift_type}->Value - $$prob2[2]{$shift_type}->Value) < $tolerance))
	  { $sum += $$prob2[1]; }
	}
      }

    push @shift_prob, [ $sum, ((exists $$prob[2]{$shift_type}) ? $$prob[2]{$shift_type} : 0) ];
    }
  
  @shift_prob = sort { $$b[0] <=> $$a[0]; } (@shift_prob);
  return (@{$shift_prob[0]});
  }

#
#
#
sub verify_shift
  {
  my $shift = shift @_;
  my $range = shift @_;

  return scalar(grep { (($shift >= $$range{$_}[0]) && ($shift <= $$range{$_}[1])); } (keys %$range));
  }

#
#
#
sub calc_distance
  {
  my $peak = shift @_;
  my $group = shift @_;
  my $tolerance_alist = shift @_;
  my $mapping = \%linear_mapping;
  if (@_)
    { $mapping = shift @_; }

  my $distance = 1000;
  foreach my $group_peak (@$group)
    {
    my $test_distance = 0;
    for(my $x=0; $x < @$tolerance_alist; $x++)
      {
      if ($$tolerance_alist[$x] != 0.0)
	{
	$test_distance += ($$group_peak[$x+1] - $$peak[$$mapping{($x+1)}]) * ($$group_peak[$x+1] - $$peak[$$mapping{($x+1)}]);
	}
      }
    if ($test_distance < $distance)
      { $distance = $test_distance; }
    }

  return sqrt($distance);
  }


# valid_grouping
#   Tests whether a peak can be joined to a peakgroup
#
# Parameters:
#	$peak1 - ref to peak 1.
#	$peakgroup - ref to array of peaks.
#	$tolerance_alist - ref to array of tolerances.
sub valid_grouping
  {
  my $peak = shift @_;
  my $peakgroup = shift @_;
  my $tolerance_alist = shift @_;
  my $mapping = \%linear_mapping;
  if (@_)
    { $mapping = shift @_; }
    

  my $peak_group_id = &group_id($peak);
  my $peak_group_id2 = &group_id($$peakgroup[0]);

  return ( (((($peak_group_id eq "") && ($peak_group_id2 eq "")) || $ignore_peak_based_overrides) && (grep { &within_tolerances($_, $peak, $tolerance_alist, $mapping); } (@$peakgroup))) ||
	  (! $ignore_peak_based_overrides && ($peak_group_id ne "") && ($peak_group_id2 ne "") && ($peak_group_id eq $peak_group_id2)));     
  }

# group_id
#   Returns any positive group id
#
# Parameters:
#	$peak - ref to peak.
#
sub group_id
  {
  my $peak = shift @_;

  if ($$peak[$#$peak] =~ /\;[Gg](\d+)/)
    { return $1; }

  return "";
  }

# within_tolerances
#   Tests whether two peaks are within tolerances
#
# Parameters:
#	$peak1 - ref to peak 1.
#	$peak2 - ref to peak 2.
#	$tolerance_alist - ref to array of tolerances.
#
sub within_tolerances
  {
  my $peak = shift @_;
  my $peak2 = shift @_;
  my $tolerance_alist = shift @_;
  my $mapping = shift @_;

  for(my $x=0; $x < @$tolerance_alist; $x++)
    {
    if (($$tolerance_alist[$x] != 0.0) && (abs($$peak[$x+1] - $$peak2[$$mapping{($x+1)}]) >= $$tolerance_alist[$x]))
      { return 0; }
    }

  return 1;
  }


# read_peak_file
#   Reads an AutoAssign peak list file and returns an array of peaks.
# 
# Parameters
#	$filename - name of peak list file.
#
sub read_peak_file 
  {
  my $filename = shift @_;
    
  local *PEAKFILE;
  if ($filename eq "-")
    { *PEAKFILE = *STDIN; }
  else
    { open(PEAKFILE, $filename) || die ("open " . $filename . " failed."); }

  my $peak_alist = [];
  my $line;
  while ($line = <PEAKFILE>) 
    {
    next if ($line =~ /^\#/); # skip comment lines
    last if ($line =~ /^[*]/); # end loop if end_of_file marker is encountered.
    
    ($line) = ($line =~/^\s*(.*)/); # skip leading white space.
    chomp $line;
    push @$peak_alist, [ split(/\s+/, $line) ];
    }

  my $comment_alist = [];
  while ($line = <PEAKFILE>)
    { push @$comment_alist, $line; }

  close PEAKFILE;
  return wantarray ? ($peak_alist, $comment_alist) : $peak_alist;
  }


